@@ -1,19 +1,17 @@
-from __future__ import division
-from __future__ import print_function
-
 import cProfile
 import gc
 import pstats
 import sys
 import time
 import traceback
+import typing
 from math import ceil
+from pathlib import Path
 
-from .compat import INT
-from .compat import XRANGE
 from .timers import compute_timer_precision
 from .utils import NameWrapper
 from .utils import format_time
+from .utils import slugify
 
 try:
     import statistics
@@ -29,11 +27,29 @@ class FixtureAlreadyUsed(Exception):
     pass
 
 
-class BenchmarkFixture(object):
-    _precisions = {}
+class BenchmarkFixture:
+    _precisions: typing.ClassVar = {}
 
-    def __init__(self, node, disable_gc, timer, min_rounds, min_time, max_time, warmup, warmup_iterations,
-                 calibration_precision, add_stats, logger, warner, disabled, cprofile, group=None):
+    def __init__(
+        self,
+        node,
+        disable_gc,
+        timer,
+        min_rounds,
+        min_time,
+        max_time,
+        warmup,
+        warmup_iterations,
+        calibration_precision,
+        add_stats,
+        logger,
+        warner,
+        disabled,
+        cprofile,
+        cprofile_loops,
+        cprofile_dump,
+        group=None,
+    ):
         self.name = node.name
         self.fullname = node._nodeid
         self.disabled = disabled
@@ -61,6 +77,8 @@ class BenchmarkFixture(object):
         self._cleanup_callbacks = []
         self._mode = None
         self.cprofile = cprofile
+        self.cprofile_loops = cprofile_loops
+        self.cprofile_dump = cprofile_dump
         self.cprofile_stats = None
         self.stats = None
 
@@ -73,9 +91,8 @@ class BenchmarkFixture(object):
             timer_precision = self._precisions[timer]
         else:
             timer_precision = self._precisions[timer] = compute_timer_precision(timer)
-            self._logger.debug("")
-            self._logger.debug("Computing precision for %s ... %ss." % (
-                NameWrapper(timer), format_time(timer_precision)), blue=True, bold=True)
+            self._logger.debug('')
+            self._logger.debug(f'Computing precision for {NameWrapper(timer)} ... {format_time(timer_precision)}s.', blue=True, bold=True)
         return timer_precision
 
     def _make_runner(self, function_to_benchmark, args, kwargs):
@@ -105,23 +122,35 @@ class BenchmarkFixture(object):
         return runner
 
     def _make_stats(self, iterations):
-        bench_stats = Metadata(self, iterations=iterations, options={
-            "disable_gc": self._disable_gc,
-            "timer": self._timer,
-            "min_rounds": self._min_rounds,
-            "max_time": self._max_time,
-            "min_time": self._min_time,
-            "warmup": self._warmup,
-        })
+        bench_stats = Metadata(
+            self,
+            iterations=iterations,
+            options={
+                'disable_gc': self._disable_gc,
+                'timer': self._timer,
+                'min_rounds': self._min_rounds,
+                'max_time': self._max_time,
+                'min_time': self._min_time,
+                'warmup': self._warmup,
+            },
+        )
         self._add_stats(bench_stats)
         self.stats = bench_stats
         return bench_stats
 
+    def _save_cprofile(self, profile: cProfile.Profile):
+        stats = pstats.Stats(profile)
+        self.stats.cprofile_stats = stats
+        if self.cprofile_dump:
+            output_file = Path(f'{self.cprofile_dump}-{slugify(self.name)}.prof')
+            output_file.parent.mkdir(parents=True, exist_ok=True)
+            stats.dump_stats(output_file)
+            self._logger.info(f'Saved profile: {output_file}', bold=True)
+
     def __call__(self, function_to_benchmark, *args, **kwargs):
         if self._mode:
             self.has_error = True
-            raise FixtureAlreadyUsed(
-                "Fixture can only be used once. Previously it was used in %s mode." % self._mode)
+            raise FixtureAlreadyUsed(f'Fixture can only be used once. Previously it was used in {self._mode} mode.')
         try:
             self._mode = 'benchmark(...)'
             return self._raw(function_to_benchmark, *args, **kwargs)
@@ -132,43 +161,50 @@ class BenchmarkFixture(object):
     def pedantic(self, target, args=(), kwargs=None, setup=None, rounds=1, warmup_rounds=0, iterations=1):
         if self._mode:
             self.has_error = True
-            raise FixtureAlreadyUsed(
-                "Fixture can only be used once. Previously it was used in %s mode." % self._mode)
+            raise FixtureAlreadyUsed(f'Fixture can only be used once. Previously it was used in {self._mode} mode.')
         try:
             self._mode = 'benchmark.pedantic(...)'
-            return self._raw_pedantic(target, args=args, kwargs=kwargs, setup=setup, rounds=rounds,
-                                      warmup_rounds=warmup_rounds, iterations=iterations)
+            return self._raw_pedantic(
+                target, args=args, kwargs=kwargs, setup=setup, rounds=rounds, warmup_rounds=warmup_rounds, iterations=iterations
+            )
         except Exception:
             self.has_error = True
             raise
 
     def _raw(self, function_to_benchmark, *args, **kwargs):
+        loops_range = None
+
         if self.enabled:
             runner = self._make_runner(function_to_benchmark, args, kwargs)
 
             duration, iterations, loops_range = self._calibrate_timer(runner)
 
-            # Choose how many time we must repeat the test
+            # Choose how many times we must repeat the test
             rounds = int(ceil(self._max_time / duration))
             rounds = max(rounds, self._min_rounds)
             rounds = min(rounds, sys.maxsize)
 
             stats = self._make_stats(iterations)
 
-            self._logger.debug("  Running %s rounds x %s iterations ..." % (rounds, iterations), yellow=True, bold=True)
+            self._logger.debug(f'  Running {rounds} rounds x {iterations} iterations ...', yellow=True, bold=True)
             run_start = time.time()
             if self._warmup:
                 warmup_rounds = min(rounds, max(1, int(self._warmup / iterations)))
-                self._logger.debug("  Warmup %s rounds x %s iterations ..." % (warmup_rounds, iterations))
-                for _ in XRANGE(warmup_rounds):
+                self._logger.debug(f'  Warmup {warmup_rounds} rounds x {iterations} iterations ...')
+                for _ in range(warmup_rounds):
                     runner(loops_range)
-            for _ in XRANGE(rounds):
+            for _ in range(rounds):
                 stats.update(runner(loops_range))
-            self._logger.debug("  Ran for %ss." % format_time(time.time() - run_start), yellow=True, bold=True)
+            self._logger.debug(f'  Ran for {format_time(time.time() - run_start)}s.', yellow=True, bold=True)
+        if self.cprofile_loops is None:
+            cprofile_loops = loops_range or range(1)
+        else:
+            cprofile_loops = range(self.cprofile_loops)
         if self.enabled and self.cprofile:
             profile = cProfile.Profile()
-            function_result = profile.runcall(function_to_benchmark, *args, **kwargs)
-            self.stats.cprofile_stats = pstats.Stats(profile)
+            for _ in cprofile_loops:
+                function_result = profile.runcall(function_to_benchmark, *args, **kwargs)
+            self._save_cprofile(profile)
         else:
             function_result = function_to_benchmark(*args, **kwargs)
         return function_result
@@ -179,14 +215,14 @@ class BenchmarkFixture(object):
 
         has_args = bool(args or kwargs)
 
-        if not isinstance(iterations, INT) or iterations < 1:
-            raise ValueError("Must have positive int for `iterations`.")
+        if not isinstance(iterations, int) or iterations < 1:
+            raise ValueError('Must have positive int for `iterations`.')
 
-        if not isinstance(rounds, INT) or rounds < 1:
-            raise ValueError("Must have positive int for `rounds`.")
+        if not isinstance(rounds, int) or rounds < 1:
+            raise ValueError('Must have positive int for `rounds`.')
 
-        if not isinstance(warmup_rounds, INT) or warmup_rounds < 0:
-            raise ValueError("Must have positive int for `warmup_rounds`.")
+        if not isinstance(warmup_rounds, int) or warmup_rounds < 0:
+            raise ValueError('Must have positive int for `warmup_rounds`.')
 
         if iterations > 1 and setup:
             raise ValueError("Can't use more than 1 `iterations` with a `setup` function.")
@@ -205,14 +241,14 @@ class BenchmarkFixture(object):
             return target(*args, **kwargs)
 
         stats = self._make_stats(iterations)
-        loops_range = XRANGE(iterations) if iterations > 1 else None
-        for _ in XRANGE(warmup_rounds):
+        loops_range = range(iterations) if iterations > 1 else None
+        for _ in range(warmup_rounds):
             args, kwargs = make_arguments()
 
             runner = self._make_runner(target, args, kwargs)
             runner(loops_range)
 
-        for _ in XRANGE(rounds):
+        for _ in range(rounds):
             args, kwargs = make_arguments()
 
             runner = self._make_runner(target, args, kwargs)
@@ -223,14 +259,21 @@ class BenchmarkFixture(object):
             stats.update(duration)
 
         if loops_range:
+            # if it has been looped then we don't have the result, we need to do 1 extra run for it
             args, kwargs = make_arguments()
             result = target(*args, **kwargs)
 
         if self.cprofile:
+            if self.cprofile_loops is None:
+                cprofile_loops = loops_range or range(1)
+            else:
+                cprofile_loops = range(self.cprofile_loops)
+
             profile = cProfile.Profile()
             args, kwargs = make_arguments()
-            profile.runcall(target, *args, **kwargs)
-            self.stats.cprofile_stats = pstats.Stats(profile)
+            for _ in cprofile_loops:
+                profile.runcall(target, *args, **kwargs)
+            self._save_cprofile(profile)
 
         return result
 
@@ -238,7 +281,7 @@ class BenchmarkFixture(object):
         try:
             import aspectlib
         except ImportError as exc:
-            raise ImportError(exc.args, "Please install aspectlib or pytest-benchmark[aspect]")
+            raise ImportError(exc.args, 'Please install aspectlib or pytest-benchmark[aspect]') from exc
 
         def aspect(function):
             def wrapper(*args, **kwargs):
@@ -255,25 +298,23 @@ class BenchmarkFixture(object):
             callback = self._cleanup_callbacks.pop()
             callback()
         if not self._mode and not self.skipped:
-            self._logger.warn("Benchmark fixture was not used at all in this test!",
-                              warner=self._warner, suspend=True)
+            self._logger.warning('Benchmark fixture was not used at all in this test!', warner=self._warner, suspend=True)
 
     def _calibrate_timer(self, runner):
         timer_precision = self._get_precision(self._timer)
         min_time = max(self._min_time, timer_precision * self._calibration_precision)
         min_time_estimate = min_time * 5 / self._calibration_precision
-        self._logger.debug("")
-        self._logger.debug("  Calibrating to target round %ss; will estimate when reaching %ss "
-                           "(using: %s, precision: %ss)." % (
-                               format_time(min_time),
-                               format_time(min_time_estimate),
-                               NameWrapper(self._timer),
-                               format_time(timer_precision)
-                           ), yellow=True, bold=True)
+        self._logger.debug('')
+        self._logger.debug(
+            f'  Calibrating to target round {format_time(min_time)}s; will estimate when reaching {format_time(min_time_estimate)}s '
+            f'(using: {NameWrapper(self._timer)}, precision: {format_time(timer_precision)}s).',
+            yellow=True,
+            bold=True,
+        )
 
         loops = 1
         while True:
-            loops_range = XRANGE(loops)
+            loops_range = range(loops)
             duration = runner(loops_range)
             if self._warmup:
                 warmup_start = time.time()
@@ -283,23 +324,20 @@ class BenchmarkFixture(object):
                     duration = min(duration, runner(loops_range))
                     warmup_rounds += 1
                     warmup_iterations += loops
-                self._logger.debug("    Warmup: %ss (%s x %s iterations)." % (
-                    format_time(time.time() - warmup_start),
-                    warmup_rounds, loops
-                ))
+                self._logger.debug(f'    Warmup: {format_time(time.time() - warmup_start)}s ({warmup_rounds} x {loops} iterations).')
 
-            self._logger.debug("    Measured %s iterations: %ss." % (loops, format_time(duration)), yellow=True)
+            self._logger.debug(f'    Measured {loops} iterations: {format_time(duration)}s.', yellow=True)
             if duration >= min_time:
                 break
 
             if duration >= min_time_estimate:
                 # coarse estimation of the number of loops
                 loops = int(ceil(min_time * loops / duration))
-                self._logger.debug("    Estimating %s iterations." % loops, green=True)
+                self._logger.debug(f'    Estimating {loops} iterations.', green=True)
                 if loops == 1:
                     # If we got a single loop then bail early - nothing to calibrate if the the
                     # test function is 100 times slower than the timer resolution.
-                    loops_range = XRANGE(loops)
+                    loops_range = range(loops)
                     break
             else:
                 loops *= 10
