@@ -1,10 +1,6 @@
-from __future__ import division
-from __future__ import print_function
-
 import argparse
 import json
 import netrc
-import ntpath
 import os
 import platform
 import re
@@ -12,6 +8,7 @@ import subprocess
 import sys
 import types
 from datetime import datetime
+from datetime import timezone
 from decimal import Decimal
 from functools import partial
 from os.path import basename
@@ -24,70 +21,55 @@ from subprocess import check_output
 from urllib.parse import parse_qs
 from urllib.parse import urlparse
 
-import genericpath
+from .compat import PY311
 
-from .compat import PY38
-
-# This is here (in the utils module) because it might be used by
-# various other modules.
-try:
-    from pathlib2 import Path  # noqa: F401
-except ImportError:
-    from pathlib import Path  # noqa: F401
-
-
-TIME_UNITS = {
-    "": "Seconds",
-    "m": "Milliseconds (ms)",
-    "u": "Microseconds (us)",
-    "n": "Nanoseconds (ns)"
-}
-ALLOWED_COLUMNS = ["min", "max", "mean", "stddev", "median", "iqr", "ops", "outliers", "rounds", "iterations"]
+TIME_UNITS = {'': 'Seconds', 'm': 'Milliseconds (ms)', 'u': 'Microseconds (us)', 'n': 'Nanoseconds (ns)'}
+ALLOWED_COLUMNS = ['min', 'max', 'mean', 'stddev', 'median', 'iqr', 'ops', 'outliers', 'rounds', 'iterations']
 
 
 class SecondsDecimal(Decimal):
     def __float__(self):
-        return float(super(SecondsDecimal, self).__str__())
+        return float(super().__str__())
 
     def __str__(self):
-        return "{0}s".format(format_time(float(super(SecondsDecimal, self).__str__())))
+        return f'{format_time(float(super().__str__()))}s'
 
     @property
     def as_string(self):
-        return super(SecondsDecimal, self).__str__()
+        return super().__str__()
 
 
-class NameWrapper(object):
+class NameWrapper:
     def __init__(self, target):
         self.target = target
 
     def __str__(self):
-        name = self.target.__module__ + "." if hasattr(self.target, '__module__') else ""
+        name = self.target.__module__ + '.' if hasattr(self.target, '__module__') else ''
         name += self.target.__name__ if hasattr(self.target, '__name__') else repr(self.target)
         return name
 
     def __repr__(self):
-        return "NameWrapper(%s)" % repr(self.target)
+        return f'NameWrapper({self.target!r})'
 
 
 def get_tag(project_name=None):
     info = get_commit_info(project_name)
     parts = [info['id'], get_current_time()]
     if info['dirty']:
-        parts.append("uncommited-changes")
-    return "_".join(parts)
+        parts.append('uncommited-changes')
+    return '_'.join(parts)
 
 
 def get_machine_id():
-    return "%s-%s-%s-%s" % (
+    return '{}-{}-{}-{}'.format(
         platform.system(),
         platform.python_implementation(),
-        ".".join(platform.python_version_tuple()[:2]),
-        platform.architecture()[0]
+        '.'.join(platform.python_version_tuple()[:2]),
+        platform.architecture()[0],
     )
 
 
-class Fallback(object):
+class Fallback:
     def __init__(self, fallback, exceptions):
         self.fallback = fallback
         self.functions = []
@@ -120,7 +102,7 @@ def get_project_name_git():
     is_git = check_output(['git', 'rev-parse', '--git-dir'], stderr=subprocess.STDOUT)
     if is_git:
         project_address = check_output(['git', 'config', '--local', 'remote.origin.url'])
-        if isinstance(project_address, bytes) and str != bytes:
+        if isinstance(project_address, bytes):
             project_address = project_address.decode()
         project_name = [i for i in re.split(r'[/:\s\\]|\.git', project_address) if i][-1]
         return project_name.strip()
@@ -131,7 +113,7 @@ def get_project_name_hg():
     with open(os.devnull, 'w') as devnull:
         project_address = check_output(['hg', 'path', 'default'], stderr=devnull)
     project_address = project_address.decode()
-    project_name = project_address.split("/")[-1]
+    project_name = project_address.split('/')[-1]
     return project_name.strip()
 
 
@@ -190,20 +172,19 @@ def get_commit_info(project_name=None):
             'time': None,
             'author_time': None,
             'dirty': dirty,
-            'error': 'CalledProcessError({0.returncode}, {0.output!r})'.format(exc)
-                     if isinstance(exc, CalledProcessError) else repr(exc),
+            'error': f'CalledProcessError({exc.returncode}, {exc.output!r})' if isinstance(exc, CalledProcessError) else repr(exc),
             'project': project_name,
             'branch': branch,
         }
 
 
 def get_current_time():
-    return datetime.utcnow().strftime("%Y%m%d_%H%M%S")
+    return datetime.now(timezone.utc).strftime('%Y%m%d_%H%M%S')
 
 
 def first_or_value(obj, value):
     if obj:
-        value, = obj
+        (value,) = obj
 
     return value
 
@@ -226,11 +207,12 @@ def short_filename(path, machine_id=None):
 
 
 def load_timer(string):
-    if "." not in string:
+    if '.' not in string:
         raise argparse.ArgumentTypeError("Value for --benchmark-timer must be in dotted form. Eg: 'module.attr'.")
-    mod, attr = string.rsplit(".", 1)
+    mod, attr = string.rsplit('.', 1)
     if mod == 'pep418':
         import time
+
         return NameWrapper(getattr(time, attr))
     else:
         __import__(mod)
@@ -238,7 +220,7 @@ def load_timer(string):
         return NameWrapper(getattr(mod, attr))
 
 
-class RegressionCheck(object):
+class RegressionCheck:
     def __init__(self, field, threshold):
         self.field = field
         self.threshold = threshold
@@ -246,16 +228,14 @@ class RegressionCheck(object):
     def fails(self, current, compared):
         val = self.compute(current, compared)
         if val > self.threshold:
-            return "Field %r has failed %s: %.9f > %.9f" % (
-                self.field, self.__class__.__name__, val, self.threshold
-            )
+            return f'Field {self.field!r} has failed {self.__class__.__name__}: {val:.9f} > {self.threshold:.9f}'
 
 
 class PercentageRegressionCheck(RegressionCheck):
     def compute(self, current, compared):
         val = compared[self.field]
         if not val:
-            return float("inf")
+            return float('inf')
         return current[self.field] / val * 100 - 100
 
 
@@ -264,10 +244,14 @@ class DifferenceRegressionCheck(RegressionCheck):
         return current[self.field] - compared[self.field]
 
 
-def parse_compare_fail(string,
-                       rex=re.compile(r'^(?P<field>min|max|mean|median|stddev|iqr):'
-                                      r'((?P<percentage>[0-9]?[0-9])%|(?P<difference>[0-9]*\.?[0-9]+([eE][-+]?['
-                                      r'0-9]+)?))$')):
+def parse_compare_fail(
+    string,
+    rex=re.compile(
+        r'^(?P<field>min|max|mean|median|stddev|iqr):'
+        r'((?P<percentage>[0-9]?[0-9])%|(?P<difference>[0-9]*\.?[0-9]+([eE][-+]?['
+        r'0-9]+)?))$'
+    ),
+):
     m = rex.match(string)
     if m:
         g = m.groupdict()
@@ -276,58 +260,71 @@ def parse_compare_fail(string,
         elif g['difference']:
             return DifferenceRegressionCheck(g['field'], float(g['difference']))
 
-    raise argparse.ArgumentTypeError("Could not parse value: %r." % string)
+    raise argparse.ArgumentTypeError(f'Could not parse value: {string!r}.')
+
+
+def parse_cprofile_loops(string):
+    if string == 'auto':
+        return None
+    else:
+        try:
+            value = int(string)
+        except ValueError:
+            raise argparse.ArgumentTypeError(f'Could not parse value: {string!r}. Expected an integer or `auto`.') from None
+        if value < 1:
+            raise argparse.ArgumentTypeError(f'Invalid value: {string!r}. Must be greater than 0.') from None
+        return value
 
 
 def parse_warmup(string):
     string = string.lower().strip()
-    if string == "auto":
-        return platform.python_implementation() == "PyPy"
-    elif string in ["off", "false", "no"]:
+    if string == 'auto':
+        return platform.python_implementation() == 'PyPy'
+    elif string in ['off', 'false', 'no']:
         return False
-    elif string in ["on", "true", "yes", ""]:
+    elif string in ['on', 'true', 'yes', '']:
         return True
     else:
-        raise argparse.ArgumentTypeError("Could not parse value: %r." % string)
+        raise argparse.ArgumentTypeError(f'Could not parse value: {string!r}.')
 
 
 def name_formatter_short(bench):
-    name = bench["name"]
-    if bench["source"]:
-        name = "%s (%.4s)" % (name, split(bench["source"])[-1])
-    if name.startswith("test_"):
+    name = bench['name']
+    if bench['source']:
+        name = '{} ({:.4})'.format(name, split(bench['source'])[-1])
+    if name.startswith('test_'):
         name = name[5:]
     return name
 
 
 def name_formatter_normal(bench):
-    name = bench["name"]
-    if bench["source"]:
-        parts = bench["source"].split('/')
+    name = bench['name']
+    if bench['source']:
+        parts = bench['source'].split('/')
         parts[-1] = parts[-1][:12]
-        name = "%s (%s)" % (name, '/'.join(parts))
+        name = '{} ({})'.format(name, '/'.join(parts))
     return name
 
 
 def name_formatter_long(bench):
-    if bench["source"]:
-        return "%(fullname)s (%(source)s)" % bench
+    if bench['source']:
+        return '{fullname} ({source})'.format(**bench)
     else:
-        return bench["fullname"]
+        return bench['fullname']
 
 
 def name_formatter_trial(bench):
-    if bench["source"]:
-        return "%.4s" % split(bench["source"])[-1]
+    if bench['source']:
+        return '{:.4}'.format(split(bench['source'])[-1])
     else:
         return '????'
 
 
 NAME_FORMATTERS = {
-    "short": name_formatter_short,
-    "normal": name_formatter_normal,
-    "long": name_formatter_long,
-    "trial": name_formatter_trial,
+    'short': name_formatter_short,
+    'normal': name_formatter_normal,
+    'long': name_formatter_long,
+    'trial': name_formatter_trial,
 }
 
 
@@ -336,7 +333,7 @@ def parse_name_format(string):
     if string in NAME_FORMATTERS:
         return string
     else:
-        raise argparse.ArgumentTypeError("Could not parse value: %r." % string)
+        raise argparse.ArgumentTypeError(f'Could not parse value: {string!r}.')
 
 
 def parse_timer(string):
@@ -345,11 +342,12 @@ def parse_timer(string):
 
 def parse_sort(string):
     string = string.lower().strip()
-    if string not in ("min", "max", "mean", "stddev", "name", "fullname"):
+    if string not in ('min', 'max', 'mean', 'stddev', 'name', 'fullname'):
         raise argparse.ArgumentTypeError(
-            "Unacceptable value: %r. "
+            f'Unacceptable value: {string!r}. '
             "Value for --benchmark-sort must be one of: 'min', 'max', 'mean', "
-            "'stddev', 'name', 'fullname'." % string)
+            "'stddev', 'name', 'fullname'."
+        )
     return string
 
 
@@ -358,8 +356,8 @@ def parse_columns(string):
     invalid = set(columns) - set(ALLOWED_COLUMNS)
     if invalid:
         # there are extra items in columns!
-        msg = "Invalid column name(s): %s. " % ', '.join(invalid)
-        msg += "The only valid column names are: %s" % ', '.join(ALLOWED_COLUMNS)
+        msg = 'Invalid column name(s): {}. '.format(', '.join(invalid))
+        msg += 'The only valid column names are: {}'.format(', '.join(ALLOWED_COLUMNS))
         raise argparse.ArgumentTypeError(msg)
     return columns
 
@@ -368,10 +366,10 @@ def parse_rounds(string):
     try:
         value = int(string)
     except ValueError as exc:
-        raise argparse.ArgumentTypeError(exc)
+        raise argparse.ArgumentTypeError(exc) from None
     else:
         if value < 1:
-            raise argparse.ArgumentTypeError("Value for --benchmark-rounds must be at least 1.")
+            raise argparse.ArgumentTypeError('Value for --benchmark-rounds must be at least 1.')
         return value
 
 
@@ -379,20 +377,19 @@ def parse_seconds(string):
     try:
         return SecondsDecimal(string).as_string
     except Exception as exc:
-        raise argparse.ArgumentTypeError("Invalid decimal value %r: %r" % (string, exc))
+        raise argparse.ArgumentTypeError(f'Invalid decimal value {string!r}: {exc!r}') from None
 
 
 def parse_save(string):
     if not string:
         raise argparse.ArgumentTypeError("Can't be empty.")
-    illegal = ''.join(c for c in r"\/:*?<>|" if c in string)
+    illegal = ''.join(c for c in r'\/:*?<>|' if c in string)
     if illegal:
-        raise argparse.ArgumentTypeError("Must not contain any of these characters: /:*?<>|\\ (it has %r)" % illegal)
+        raise argparse.ArgumentTypeError(f'Must not contain any of these characters: /:*?<>|\\ (it has {illegal!r})')
     return string
 
 
 def _parse_hosts(storage_url, netrc_file):
-
     # load creds from netrc file
     path = os.path.expanduser(netrc_file)
     creds = None
@@ -402,84 +399,81 @@ def _parse_hosts(storage_url, netrc_file):
     # add creds to urls
     urls = []
     for netloc in storage_url.netloc.split(','):
-        auth = ""
+        auth = ''
         if creds and '@' not in netloc:
             host = netloc.split(':').pop(0)
             res = creds.authenticators(host)
             if res:
                 user, _, secret = res
-                auth = "{user}:{secret}@".format(user=user, secret=secret)
-        url = "{scheme}://{auth}{netloc}".format(scheme=storage_url.scheme,
-                                                 netloc=netloc, auth=auth)
+                auth = f'{user}:{secret}@'
+        url = f'{storage_url.scheme}://{auth}{netloc}'
         urls.append(url)
     return urls
 
 
-def parse_elasticsearch_storage(string, default_index="benchmark",
-                                default_doctype="benchmark", netrc_file=''):
+def parse_elasticsearch_storage(string, default_index='benchmark', default_doctype='benchmark', netrc_file=''):
     storage_url = urlparse(string)
     hosts = _parse_hosts(storage_url, netrc_file)
     index = default_index
     doctype = default_doctype
-    if storage_url.path and storage_url.path != "/":
-        splitted = storage_url.path.strip("/").split("/")
+    if storage_url.path and storage_url.path != '/':
+        splitted = storage_url.path.strip('/').split('/')
         index = splitted[0]
         if len(splitted) >= 2:
             doctype = splitted[1]
     query = parse_qs(storage_url.query)
     try:
-        project_name = query["project_name"][0]
+        project_name = query['project_name'][0]
     except KeyError:
         project_name = get_project_name()
     return hosts, index, doctype, project_name
 
 
 def load_storage(storage, **kwargs):
-    if "://" not in storage:
-        storage = "file://" + storage
+    if '://' not in storage:
+        storage = 'file://' + storage
     netrc_file = kwargs.pop('netrc')  # only used by elasticsearch storage
-    if storage.startswith("file://"):
+    if storage.startswith('file://'):
         from .storage.file import FileStorage
-        return FileStorage(storage[len("file://"):], **kwargs)
-    elif storage.startswith("elasticsearch+"):
+
+        return FileStorage(storage[len('file://') :], **kwargs)
+    elif storage.startswith('elasticsearch+'):
         from .storage.elasticsearch import ElasticsearchStorage
 
         # TODO update benchmark_autosave
-        args = parse_elasticsearch_storage(storage[len("elasticsearch+"):],
-                                           netrc_file=netrc_file)
+        args = parse_elasticsearch_storage(storage[len('elasticsearch+') :], netrc_file=netrc_file)
         return ElasticsearchStorage(*args, **kwargs)
     else:
-        raise argparse.ArgumentTypeError("Storage must be in form of file://path or "
-                                         "elasticsearch+http[s]://host1,host2/index/doctype")
+        raise argparse.ArgumentTypeError('Storage must be in form of file://path or ' 'elasticsearch+http[s]://host1,host2/index/doctype')
 
 
 def time_unit(value):
     if value < 1e-6:
-        return "n", 1e9
+        return 'n', 1e9
     elif value < 1e-3:
-        return "u", 1e6
+        return 'u', 1e6
     elif value < 1:
-        return "m", 1e3
+        return 'm', 1e3
     else:
-        return "", 1.
+        return '', 1.0
 
 
 def operations_unit(value):
-    if value > 1e+6:
-        return "M", 1e-6
-    if value > 1e+3:
-        return "K", 1e-3
-    return "", 1.
+    if value > 1e6:
+        return 'M', 1e-6
+    if value > 1e3:
+        return 'K', 1e-3
+    return '', 1.0
 
 
 def format_time(value):
     unit, adjustment = time_unit(value)
-    return "{0:.2f}{1:s}".format(value * adjustment, unit)
+    return f'{value * adjustment:.2f}{unit:s}'
 
 
-class cached_property(object):
+class cached_property:
     def __init__(self, func):
-        self.__doc__ = getattr(func, '__doc__')
+        self.__doc__ = func.__doc__
         self.func = func
 
     def __get__(self, obj, cls):
@@ -516,11 +510,27 @@ def clonefunc(f):
     if not hasattr(f, '__code__'):
         return f
     co = f.__code__
-    args = [co.co_argcount, co.co_kwonlyargcount, co.co_nlocals, co.co_stacksize, co.co_flags, co.co_code,
-            co.co_consts, co.co_names, co.co_varnames, co.co_filename, co.co_name,
-            co.co_firstlineno, co.co_lnotab, co.co_freevars, co.co_cellvars]
-    if PY38:
-        args.insert(1, co.co_posonlyargcount)
+    args = [
+        co.co_argcount,
+        co.co_posonlyargcount,
+        co.co_kwonlyargcount,
+        co.co_nlocals,
+        co.co_stacksize,
+        co.co_flags,
+        co.co_code,
+        co.co_consts,
+        co.co_names,
+        co.co_varnames,
+        co.co_filename,
+        co.co_name,
+        co.co_firstlineno,
+        co.co_lnotab,
+        co.co_freevars,
+        co.co_cellvars,
+    ]
+    if PY311:
+        args.insert(12, co.co_qualname)
+        args.insert(15, co.co_exceptiontable)
     co2 = types.CodeType(*args)
     #
     # then, we clone the function itself, using the new co2
@@ -528,13 +538,13 @@ def clonefunc(f):
     return f2
 
 
-def format_dict(obj):
-    return "{%s}" % ", ".join("%s: %s" % (k, json.dumps(v)) for k, v in sorted(obj.items()))
-
-
 class SafeJSONEncoder(json.JSONEncoder):
     def default(self, o):
-        return "UNSERIALIZABLE[%r]" % o
+        return f'UNSERIALIZABLE[{o!r}]'
+
+
+def consistent_dumps(value):
+    return json.dumps(value, sort_keys=True)
 
 
 def safe_dumps(obj, **kwargs):
@@ -554,71 +564,20 @@ def report_progress(iterable, terminal_reporter, format_string, **kwargs):
 
 
 def report_noprogress(iterable, *args, **kwargs):
-    for pos, item in enumerate(iterable):
-        yield "", item
+    for item in iterable:
+        yield '', item
 
 
 def report_online_progress(progress_reporter, tr, line):
-    next(progress_reporter([line], tr, "{value}"))
+    next(progress_reporter([line], tr, '{value}'))
 
 
 def slugify(name):
-    for c in r"\/:*?<>| ":
+    for c in r'\/:*?<>| ':
         name = name.replace(c, '_').replace('__', '_')
     return name
 
 
-def commonpath(paths):
-    """Given a sequence of path names, returns the longest common sub-path."""
-
-    if not paths:
-        raise ValueError('commonpath() arg is an empty sequence')
-
-    if isinstance(paths[0], bytes):
-        sep = b'\\'
-        altsep = b'/'
-        curdir = b'.'
-    else:
-        sep = '\\'
-        altsep = '/'
-        curdir = '.'
-
-    try:
-        drivesplits = [ntpath.splitdrive(p.replace(altsep, sep).lower()) for p in paths]
-        split_paths = [p.split(sep) for d, p in drivesplits]
-
-        try:
-            isabs, = set(p[:1] == sep for d, p in drivesplits)
-        except ValueError:
-            raise ValueError("Can't mix absolute and relative paths")
-
-        # Check that all drive letters or UNC paths match. The check is made only
-        # now otherwise type errors for mixing strings and bytes would not be
-        # caught.
-        if len(set(d for d, p in drivesplits)) != 1:
-            raise ValueError("Paths don't have the same drive")
-
-        drive, path = ntpath.splitdrive(paths[0].replace(altsep, sep))
-        common = path.split(sep)
-        common = [c for c in common if c and c != curdir]
-
-        split_paths = [[c for c in s if c and c != curdir] for s in split_paths]
-        s1 = min(split_paths)
-        s2 = max(split_paths)
-        for i, c in enumerate(s1):
-            if c != s2[i]:
-                common = common[:i]
-                break
-        else:
-            common = common[:len(s1)]
-
-        prefix = drive + sep if isabs else drive
-        return prefix + sep.join(common)
-    except (TypeError, AttributeError):
-        genericpath._check_arg_types('commonpath', *paths)
-        raise
-
-
 def get_cprofile_functions(stats):
     """
     Convert pstats structure to list of sorted dicts about each function.
@@ -630,21 +589,27 @@ def get_cprofile_functions(stats):
     for function_info, run_info in stats.stats.items():
         file_path = function_info[0]
         if file_path.startswith(project_dir_parent):
-            file_path = file_path[len(project_dir_parent):].lstrip('/')
-        function_name = '{0}:{1}({2})'.format(file_path, function_info[1], function_info[2])
+            file_path = file_path[len(project_dir_parent) :].lstrip('/')
+        function_name = f'{file_path}:{function_info[1]}({function_info[2]})'
 
-        # if the function is recursive write number of 'total calls/primitive calls'
-        if run_info[0] == run_info[1]:
-            calls = str(run_info[0])
+        pcalls, ncalls, tottime, cumtime = run_info[:4]
+
+        # if the function is recursive, write number of 'total calls/primitive calls'
+        if pcalls == ncalls:
+            calls = str(pcalls)
         else:
-            calls = '{1}/{0}'.format(run_info[0], run_info[1])
+            calls = f'{ncalls}/{pcalls}'
 
-        result.append(dict(ncalls_recursion=calls,
-                           ncalls=run_info[1],
-                           tottime=run_info[2],
-                           tottime_per=run_info[2] / run_info[0] if run_info[0] > 0 else 0,
-                           cumtime=run_info[3],
-                           cumtime_per=run_info[3] / run_info[0] if run_info[0] > 0 else 0,
-                           function_name=function_name))
+        result.append(
+            {
+                'ncalls_recursion': calls,
+                'ncalls': ncalls,
+                'tottime': tottime,
+                'tottime_per': tottime / pcalls if pcalls > 0 else 0,
+                'cumtime': cumtime,
+                'cumtime_per': cumtime / pcalls if pcalls > 0 else 0,
+                'function_name': function_name,
+            }
+        )
 
     return result
